'''Driver for AWS S3 compute'''

# Libs import
from basepair.modules.aws import S3

# App import
from .abstract import StorageAbstract

FILE_NOT_FOUND = 'file_not_found'
RESTORE_COMPLETE = 'restore_complete'
RESTORE_ERROR = 'restore_error'
RESTORE_IN_PROGRESS = 'restore_in_progress'
RESTORE_NOT_REQUIRED = 'restore_not_required'
RESTORE_NOT_STARTED = 'restore_not_started'

class Driver(StorageAbstract):
  '''AWS S3 Driver class'''

  def __init__(self, cfg=None): # pylint: disable=super-init-not-called
    '''Instance constructor'''
    self.storage_settings = cfg.get('settings', {})
    self.s3_service = S3({
      'bucket': self.storage_settings.get('bucket'),
      'credentials': cfg.get('credentials'),
      'region': self.storage_settings.get('region'),
      'endpoint_url': self.storage_settings.get('endpoint_url'),
      'disable_sts': self.storage_settings.get('disable_sts', False),
    })

  def bulk_delete(self, uris):
    '''Delete list of files by their uris'''
    return self.s3_service.bulk_delete(uris)

  def delete(self, uri):
    '''Delete file from storage'''
    key = S3.get_key_from_uri(uri)
    return self.s3_service.delete(key)

  def download(self, uri, callback=None, file=None):
    '''Download file froms torage'''
    bucket = S3.get_bucket_from_uri(uri)
    key = S3.get_key_from_uri(uri)
    if callback:
      return self.s3_service.client.download_fileobj(
        bucket,
        key,
        file,
        Callback=callback,
      )
    return self.s3_service.download(key, file)

  def get_body(self, uri):
    '''Get file body'''
    key = S3.get_key_from_uri(uri)
    return self.s3_service.get_file_body(key)

  def get_head(self, uri):
    '''Get file head'''
    key = S3.get_key_from_uri(uri)
    return self.s3_service.get_object_head(key)

  def get_lifecycle(self, bucket=None):
    '''Get storage lifecycle'''
    return self.s3_service.get_lifecycle(bucket)

  def get_overall_status(self, uris):
    '''Get overall sample files status'''
    restore_statuses = [self.get_status(uri) for uri in uris]

    # Check if all statuses are RESTORE_COMPLETE
    if all(status == RESTORE_COMPLETE for status in restore_statuses):
      return RESTORE_COMPLETE
    # Find the first status
    for state in [FILE_NOT_FOUND, RESTORE_IN_PROGRESS, RESTORE_ERROR, RESTORE_NOT_STARTED]:
      if any(status == state for status in restore_statuses):
        return state
    # If no status found, return RESTORE_NOT_REQUIRED
    return RESTORE_NOT_REQUIRED

  def get_public_url(self, uri):
    '''Get a public accessible url'''
    key = S3.get_key_from_uri(uri)
    return self.s3_service.get_self_signed(key)

  def get_service(self):
    return self.s3_service

  def get_storage_context(self):
    '''Get the storage context'''
    return self.s3_service.get_storage_context()

  def get_status(self, uri):
    '''Get the file status'''
    bucket = S3.get_bucket_from_uri(uri)
    key = S3.get_key_from_uri(uri)
    return self.s3_service.get_status(bucket=bucket, key=key)

  def get_uri(self, key):
    '''Get uri using key and storage settings'''
    return f's3://{self.s3_service.bucket}/{key}'

  def list(self, prefix, bucket=None):
    return self.s3_service.list(prefix, bucket)

  def restore_files_from_cold(self, uris, days):
    '''Restore files from cold storage'''
    for uri in uris:
      status = self.get_status(uri)
      if status == 'restore_not_started':
        self.restore_from_cold(uri, days=days)

  def restore_from_cold(self, uri, days):
    '''Restore file from cold storage'''
    key = S3.get_key_from_uri(uri)
    return self.s3_service.start_restore(key, days)

  def set_body(self, body, uri):
    '''Set file body'''
    key = S3.get_key_from_uri(uri)
    return self.s3_service.set_file_body(body, key)

  def set_lifecycle(self, **kwargs):
    '''Set storage lifecycle'''
    return self.s3_service.set_lifecycle(**kwargs)

  def upload(self, file_name, full_path, **kwargs):
    '''Upload file to storage'''
    return self.s3_service.upload_file(file_name, full_path, **kwargs)



'''Driver for AWS Health Omics'''

# Libs import
from basepair.modules.aws import HOS

# App import
from .aws_s3 import Driver as S3Driver

class Driver(S3Driver):
  '''AWS Health Omics Driver class'''

  def __init__(self, cfg=None):  # pylint: disable=super-init-not-called
    '''Instance constructor'''
    super().__init__(cfg=cfg)  # calling parent class(S3) constructor to create the S3 client

    storage_settings = cfg.get('settings', {})
    self.hos_service = HOS({
      'credentials': cfg.get('credentials'),
      'sequence_store_id': storage_settings.get('ho_sequence_store_id'),
      'region': storage_settings.get('ho_region'),
      'role_arn': storage_settings.get('ho_import_export_role_arn'),
    })

  def get_service(self):
    return self.hos_service



'''Abstract class for storage drivers'''

def raise_no_implemented(msg='Abstract method require to be implemented.'):
  '''Helper to raise exception when call no implemented method'''
  raise Exception(msg)

# pylint: disable=no-self-use,unused-argument
class StorageAbstract:
  '''Abstract Driver class'''

  def __init__(self, cfg=None):
    '''Instance constructor'''
    raise Exception('Abstract class cannot be instantiated.')

  def bulk_delete(self, uris):
    '''Delete list of files by their uris'''
    raise_no_implemented()

  def delete(self, uri):
    '''Delete file from storage'''
    raise_no_implemented()

  def download(self, uri, callback=None, file=None):
    '''Download file from storage'''
    raise_no_implemented()

  def get_body(self, uri):
    '''Get file body'''
    raise_no_implemented()

  def get_head(self, uri):
    '''Get file head'''
    raise_no_implemented()

  def get_lifecycle(self, bucket=None):
    '''Get storage lifecycle'''
    raise_no_implemented()

  def get_overall_status(self, uris):
    '''Get overall sample files status'''
    raise_no_implemented()

  def get_public_url(self, uri):
    '''Get a public accessible url'''
    raise_no_implemented()

  def get_service(self):
    '''Get storage service object'''
    raise_no_implemented()

  def get_status(self, uri):
    '''Get the file status'''
    raise_no_implemented()

  def get_uri(self, key):
    '''Get uri using key and storage settings'''
    raise_no_implemented()

  def list(self, prefix, bucket=None):
    '''List files in prefix'''
    raise_no_implemented()

  def restore_files_from_cold(self, uris, days):
    '''Restore files from cold storage'''
    raise_no_implemented()

  def restore_from_cold(self, uri, days):
    '''Restore file from cold storage'''
    raise_no_implemented()

  def set_body(self, body, uri):
    '''Set file body'''
    raise_no_implemented()

  def set_lifecycle(self, **kwargs):
    '''Set storage lifecycle'''
    raise_no_implemented()

  def upload(self, file_name, full_path, force=False):
    '''Upload file to storage'''
    raise_no_implemented()

  def get_storage_context(self):
    '''Get the storage context according to driver'''
    raise_no_implemented()



'''Driver for AWS S3 compute'''

# App import
from .aws_s3 import Driver as S3Driver
from .abstract import raise_no_implemented

# constants

class Driver(S3Driver):
  '''Min IO Driver class'''

  def __init__(self, cfg=None):
    super().__init__(cfg)

  def get_lifecycle(self, bucket=None):
    raise_no_implemented("MinIO does not support lifecycle")

  def get_storage_context(self):
    '''Get the context for Minio'''
    return {
      **super().get_storage_context(),
      'storage_url' : self.storage_settings.get('endpoint_url'),
      'storage_sse_enabled' : 'False'
    }

  def restore_from_cold(self, uri, days):
    raise_no_implemented("MinIO does not support restore from cold")

  def set_lifecycle(self, **kwargs):
    raise_no_implemented("MinIO does not support lifecycle")



'''AWS HealthOmics Storage Wrappers'''

# General imports
import uuid

# Libs imports
from botocore.client import Config
from botocore.exceptions import ClientError

# Module imports
from basepair.modules.aws.service import Service

class HOS(Service):
  '''Wrapper for Health Omics Storage services'''

  def __init__(self, cfg):
    super().__init__(cfg, 'HOS')
    self.client = self.session.client(**{
      'config': Config(retries={'max_attempts': 0, 'mode': 'standard'}),
      'service_name': 'omics',
    })
    self.sequence_store_id = cfg.get('sequence_store_id')
    self.reference_store_id = cfg.get('reference_store_id')
    self.role_arn = cfg.get('role_arn')

  def get_export_job(self, job_id):
    '''Get export job'''
    try:
      response = self.client.get_read_set_export_job(
        id=job_id,
        sequenceStoreId=self.sequence_store_id,
      )
    except ClientError as error:
      self.get_log_msg({
        'exception': error,
        'msg': f'Not able to get HealthOmics export job information: {str(error)}.',
      })
      raise error
    return response

  def get_read_set_import_job(self, job_id):
    '''Get import job'''
    try:
      response = self.client.get_read_set_import_job(
        id=job_id,
        sequenceStoreId=self.sequence_store_id,
      )
    except ClientError as error:
      self.get_log_msg({
        'exception': error,
        'msg': f'Not able to get HealthOmics import job information: {str(error)}.',
      })
      raise error
    return response

  def get_reference_import_job(self, job_id):
    '''Get reference import job'''
    try:
      response = self.client.get_reference_import_job(
        id=job_id,
        referenceStoreId=self.reference_store_id,
      )
    except ClientError as error:
      self.get_log_msg({
        'exception': error,
        'msg': f'Not able to get HealthOmics reference import job information: {str(error)}.',
      })
      raise error
    return response

  def list_read_sets(self, filters):
    '''Get read sets list'''
    try:
      response = self.client.list_read_sets(
        sequenceStoreId=self.sequence_store_id,
        filter=filters
      )
    except ClientError as error:
      self.get_log_msg({
        'exception': error,
        'msg': f'Not able to list HealthOmics read sets: {str(error)}.',
      })
      raise error
    return response

  def delete_read_sets(self, read_set_ids):
    '''Delete read sets'''
    try:
      response = self.client.batch_delete_read_set(
        sequenceStoreId=self.sequence_store_id,
        ids=read_set_ids
      )
    except ClientError as error:
      self.get_log_msg({
        'exception': error,
        'msg': f'Not able to delete HealthOmics read sets: {str(error)}.',
      })
      raise error
    return response

  def start_export_job(self, destination, sources):
    '''Start export job'''
    try:
      client_token = str(uuid.uuid4())
      response = self.client.start_read_set_export_job(
        sequenceStoreId=self.sequence_store_id,
        destination=destination,
        roleArn=self.role_arn,
        clientToken=client_token,
        sources=sources
      )
    except ClientError as error:
      self.get_log_msg({
        'exception': error,
        'msg': f'Not able to start HealthOmics export job: {str(error)}.',
      })
      raise error
    return response

  def start_import_job(self, sources):
    '''Start import job'''
    try:
      client_token = str(uuid.uuid4())
      return self.client.start_read_set_import_job(
        sequenceStoreId=self.sequence_store_id,
        roleArn=self.role_arn,
        clientToken=client_token,
        sources=sources
      )
    except ClientError as error:
      self.get_log_msg({
        'exception': str(error),
        'msg': f'Not able to start HealthOmics import job: {str(error)}.',
      })
      raise error

  def start_reference_import_job(self, sources):
    '''Start reference import job'''
    try:
      client_token = str(uuid.uuid4())
      return self.client.start_reference_import_job(
        referenceStoreId=self.reference_store_id,
        roleArn=self.role_arn,
        clientToken=client_token,
        sources=sources
      )
    except ClientError as error:
      self.get_log_msg({
        'exception': str(error),
        'msg': f'Not able to start HealthOmics reference import job: {str(error)}.',
      })
      raise error